/*
 * OpenFDL Compiler
 * Copyright (C) 2005 Matt Acosta
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301, USA.
 */

package org.openfdl.core;

import org.openfdl.text.FDLTreeParserTokenTypes;

/**
 * FrameAttrNode.java
 * 
 * This class is primarily used to create a simple linked list that 
 * simplifies the output of frame attribute data generated from the *.fdl 
 * file.
 * 
 * @author Matt Acosta
 * @version 1.00
 */
public class FrameAttrNode implements FDLTreeParserTokenTypes
{
	private int attrType;
	private int slotNum;
	private String attrName;
	private String params;
	private String varType;
	
	private FrameAttrNode next;
	
	private final String newline = System.getProperty("line.separator");
	
	/*
	 * CONSTRUCTORS
	 */
	
	/**
	 * Creates an empty object. You will have to use the set methods in this 
	 * class this initialize the data fields prior to adding it to the list.
	 */
	public FrameAttrNode()
	{
		next = null;
	}
	
	/**
	 * Creates an object with the specified values already set.
	 * @param attrType Type of node we are creating. This is a value 
	 * representing "var", "appvar", "func" or "appfunc" from the <code>
	 * FDLTreeParserTokenTypes</code> interface. (Prefixed with "LITERAL_")
	 * @param slotNum Attribute slot number. This number must be greater than 
	 * 0x10000 (OO_FIRST_APP_ATTR) since the rest are reserved by the AMS 
	 * operating system.
	 * @param attrName Name of the given attribute, either the variable name 
	 * or function name.
	 * @param params Parameters for a function, this value may be null for 
	 * variables since it is not used in that case.
	 * @param varType The variable type or function return type, depending.
	 */
	public FrameAttrNode(int attrType, int slotNum, String attrName,
		String params, String varType)
	{
		this.attrType = attrType;
		this.slotNum = slotNum;
		this.attrName = attrName;
		this.params = params;
		this.varType = varType;
		
		next = null;
	}
	
	/*
	 * NODE WALKING
	 */
	
	/**
	 * Appends a new <code>FrameAttrNode</code> to this node. For more info, 
	 * see the links below.
	 * @see #FrameAttrNode(int, int, String, String, String)
	 * @see #appendAttrNode(FrameAttrNode)
	 */
	public FrameAttrNode appendAttrNode(int attrType, int slotNum,
		String attrName, String params, String varType)
	{
		return appendAttrNode(new FrameAttrNode(attrType, slotNum, attrName,
			params, varType));
	}
	
	/**
	 * This methods sets the <code>next</code> field to the parameter value 
	 * given, thereby adding it on to the linked list we are creating.
	 * @param node <code>FrameAttrNode</code> object to append to the list.
	 * @return The next node in the list, meaning, this method returns what 
	 * you gave it, the <code>node</code> parameter.
	 * @see #appendAttrNode(int, int, String, String, String)
	 */
	public FrameAttrNode appendAttrNode(FrameAttrNode node)
	{
		return next = node;
	}
	
	/*
	 * NODE -> HEADER
	 */
	
	/**
	 * Only for this node, will create a string of data that can be used in 
	 * a header file that can then be used in a <code>FRAME</code> macro when 
	 * creating a flash application. You will generally have to call this 
	 * multiple times along with the <code>getNext()</code> method.
	 * @return String representing all the data that will be output to the 
	 * header file, for this particular node.
	 * @see #getNext()
	 */
	public String createHeader()
	{
		if (attrType == LITERAL_var || attrType == LITERAL_appvar)
			return getVarString();
		else
			return getFuncString();
	}
	
	/**
	 * From the class fields, creates all necessary variable data which 
	 * includes a symbolic name and get/set macros.
	 * @return String with data that can be used with a C compiler.
	 */
	private String getVarString()
	{
		// Comment
		String outStr = "/* " + varType + " " + attrName + " */" + newline;
		
		// Symbolic name
		outStr += "#define OO_" + attrName.toUpperCase() + " (" + slotNum +
			")" + newline;
		
		// Get macro
		outStr += "#define Get" + attrName + "(obj) \\" + newline +
			"\t(" + varType + ")OO_GetAppAttr(obj, " + slotNum + ")" +
			newline;
		
		// Set macro
		outStr += "#define Set" + attrName + "(obj, value) \\" + newline +
			"\tOO_SetAppAttr(obj, " + slotNum + ", (void *)value)";
		
		return outStr;
	}
	
	/**
	 * Generates header code for use with a func or appfunc type. Data 
	 * includes a symbolic name and method call macro.
	 * @return String with data that can be used with a C compiler.
	 */
	private String getFuncString()
	{
		String pSumm = getParamSummary();
		
		// Comment
		String outStr = "/* " + varType + " " + attrName + "(" + params +
			") */" + newline;
		
		// Symbolic name
		outStr += "#define OO_" + attrName.toUpperCase() + " (" + slotNum +
			") " + newline;
		
		// Method call macro
		outStr += "#define " + attrName + "(obj" + pSumm + ") \\" + newline;
		outStr += "\t((" + varType + "(* const)(" + params +
			"))OO_GetAppAttr(obj," + slotNum + "))(obj" + pSumm + ")";
		
		return outStr;
	}
	
	/**
	 * From the parameter string, creates a "summary" for use with a macro. 
	 * More specifically, for the second parameter to a function (since the 
	 * first parameter is always "obj") summarizes to a single letter 'a'. 
	 * The third would then be 'b', the fourth 'c' and so on. If the 
	 * parameter string is null you get an empty string, and if there are 
	 * more than 26 parameters, you get "a", so too bad. 
	 * @return String with the ABCs; separated by commas.
	 * @see #getFuncString()
	 */
	private String getParamSummary()
	{
		String tempStr = "";
		
		if (params == null)
			return tempStr;
		
		int numParams = params.split(",").length - 1;
		
		if (numParams > 26) // WHY?!
		{
			System.err.println("warning: Too many parameters, setting to 1");
			numParams = 1;
		}
		
		for (int i = 0; i < numParams; i++)
			tempStr += "," + (char)(i + 'a');
		
		return tempStr;
	}
	
	/*
	 *  OVERRIDES
	 */
	
	public String toString()
	{
		return "type:" + attrType + ", slotNum:" + slotNum + ", attrName=" +
			attrName + ", params=[" + params + "], varType=" + varType;
	}
	
	/*
	 * GETTERS & SETTERS
	 */
	
	public FrameAttrNode getNext() {
		return next;
	}

	public void setAttrName(String attrName) {
		this.attrName = attrName;
	}

	public void setAttrType(int attrType) {
		this.attrType = attrType;
	}

	public void setParams(String params) {
		this.params = params;
	}

	public void setSlotNum(int slotNum) {
		this.slotNum = slotNum;
	}

	public void setVarType(String varType) {
		this.varType = varType;
	}
}
