/*
 * OpenFDL Compiler
 * Copyright (C) 2005 Matt Acosta
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301, USA.
 */

package org.openfdl.core;

import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintStream;
import java.util.Calendar;

import org.openfdl.text.FDLParser;
import org.openfdl.text.FDLScanner;
import org.openfdl.text.FDLTreeParser;

import antlr.RecognitionException;
import antlr.TokenStreamException;

/**
 * OpenFDL.java
 * 
 * Main entry point for OpenFDL. Meaning this is where arguments are checked, 
 * the parser is called and output processing is handled. Also contains the 
 * program name and version strings in case someone cares.
 * 
 * @author Matt Acosta
 * @version 1.00
 */
public class OpenFDL
{
	protected static final String PRODUCT_NAME = "OpenFDL Compiler";
	protected static final String PRODUCT_VERSION = "0.9.0";
	
	private static boolean buildStatus = true;
	
	/**
	 * Checks areguments and calls the parser.
	 * @param args Program arguments from the console. Possible arguments 
	 * include: 
	 * <ul>
	 *   <li>inputfile.fdl</li>
	 *   <li>outputfile.h - Optional</li>
	 *   <li>"?" or "help"</li>
	 * </ul>
	 * @see #parseFile(String, String)
	 * @see #printUsage(PrintStream)
	 */
	public static void main(String[] args)
	{
		if (args.length < 1)
		{
			printUsage(System.out);
			System.exit(0);
		}
		
		if (args[0].equals("?") || args[0].equals("help"))
		{
			printUsage(System.out);
			System.exit(0);
		}
		
		String outFile;
		
		// Get output file
		if (args.length >= 2)
			outFile = args[1];
		else
		{
			int extIdx = args[0].lastIndexOf(".fdl");
			
			if (extIdx == -1)		// Odd, no extension
				outFile = args[0] + ".h";
			else
				outFile = args[0].substring(0, extIdx) + ".h";
		}
		
		// Process file
		parseFile(args[0], outFile);
	}
	
	/**
	 * Changes the build status to failure. It's pretty hard to really 
	 * necesitate this, besides a failure opening a input/output file.
	 *
	 */
	public static void setBuildFailed()
	{
		buildStatus = false;
	}
	
	/**
	 * Parses an input file to create a properly formatted C header file. 
	 * First checks input file grammer, outputs the file header, then 
	 * traverses the AST tree to create and output frame attribute nodes.
	 * @param inFile Create output from this file.
	 * @param outFile Header file to send frame attribute data to.
	 * @see FrameAttrNode
	 */
	private static void parseFile(String inFile, String outFile)
	{
		FileReader fReader = null;
		PrintStream fWriter = null;
		
		try
		{
			fReader = new FileReader(inFile);
			fWriter = new PrintStream(outFile);
			
			System.err.println("Processing..");
			
			// Start grammer check
			FDLScanner scanner = new FDLScanner(fReader);
			FDLParser parser = new FDLParser(scanner);
			
			parser.parse();
			
			// Begin output
			printFileHeader(fWriter, inFile, outFile);
			
			FDLTreeParser treeParser = new FDLTreeParser();
			treeParser.parse(parser.getAST());
			
			// Walk the tree
			FrameAttrNode node = treeParser.getRootNode();
			while (node != null)
			{
				fWriter.println(node.createHeader());
				fWriter.println();
				
				node = node.getNext();
			}
		}
		catch (FileNotFoundException e)
		{
			System.err.println("Input file \"" + inFile +
				"\" could not be found.");
			
			System.err.println("Build Failed");
			
			System.exit(1);
		}
		catch (TokenStreamException e)
		{
			System.err.println("TokenStreamException, " + e.getMessage());
			
			System.err.println("Build Failed");
			
			System.exit(1);
		}
		catch (RecognitionException e)
		{
			System.err.println("RecognitionException");
			System.err.println(e.getFilename() + ", line " + e.getLine() +
				", " + e.getMessage());
			
			System.err.println("Build Failed");
			
			System.exit(1);
		}
		finally
		{
			try
			{
				if (fReader != null)
					fReader.close();
				
				if (fWriter != null)
					fWriter.close();
			}
			catch (IOException e)
			{
				System.err.println("Unable to close \"" + inFile + "\"");
				System.exit(1);
			}
		}
		
		System.err.println("Build " + (buildStatus ?
			"Successful" : "Failed"));
	}
	
	/**
	 * Prints the file header to a supplied <code>PrintStream</code> object. 
	 * More specifically just outputs the file we are to create, the date, 
	 * the input we got, and what is outputing this code (this compiler).
	 * @param out <code>PrintStream</code> to output this info to. Usually a 
	 * file.
	 * @param inFile The input filename that we are processing.
	 * @param outFile The output filename.
	 */
	private static void printFileHeader(PrintStream out, String inFile,
		String outFile)
	{
		out.println("/*");
		out.println(format("FILE:", 12) + " " + outFile);
		out.println(format("CREATED:", 12) + " " +
			Calendar.getInstance().getTime());
		out.println(format("INPUT:", 12) + " " + inFile);
		out.println(format("GENERATOR:", 12) + " " + PRODUCT_NAME +
			", v" + PRODUCT_VERSION);
		out.println("*/");
		out.println();
	}
	
	/**
	 * Prints help on how to run this program from the console. 
	 * <code>java -jar openfdl.jar inputfile.fdl [outputfile.h]</code> 
	 * <code>inputfile.fdl</code> is a *.fdl file created by the user. 
	 * <code>outputfile.h</code> is the C header file to send output. If 
	 * help is needed you may also use <code>?</code> or <code>help</code>.
	 * @param out Send output to this <code>PrintStream</code> object. 
	 * Usually <code>System.out</code>. 
	 */
	private static void printUsage(PrintStream out)
	{
		out.print("Usage of OpenFDL Compiler\n" +
			"java -jar openfdl.jar inputfile.fdl [outputfile.h]\n\n" +
			"Flags:\n" +
			"\tinputfile.fdl\tName of the file being processed.\n" +
			"\t[outputfile.h]\tOptional, name of the file to be created.\n" +
			"\t\t\tIf not specified, inputfile.h will be used.\n" +
			"\t['?' or 'help'] To display this message.. or do nothing.");
	}
	
	/**
	 * Right justifies the input string over the given width. If the given 
	 * width is lower than the length of the string, nothing happens. 
	 * Otherwise, the extra width will be added to the left of the string. 
	 * <br>Ex: <code>System.out.println(format("OpenFDL", 10));</code>  
	 * results in [   OpenFDL] being printed.
	 * @param str Input string to right justify.
	 * @param width Length of new string to create, generally greater than 
	 * length of input string.
	 * @return String containing enough spaces to right justify the given 
	 * string over the input width.
	 */
	private static String format(String str, int width)
	{
		String spaces = "";
		int padding = width - str.length();
		
		while (padding > 0)
		{
			spaces += " ";
			padding--;
		}
		
		return spaces + str;
	}
}
